@@ -123,14 +123,14 @@ void write_plt_header<E>(Context<E> &ctx, u8 *buf) {
   write_itype(buf + 16, gotplt - plt);
 }
 
-constexpr ul32 plt_entry_64[] = {
+static constexpr ul32 plt_entry_64[] = {
   0x0000'0e17, // auipc   t3, %pcrel_hi(function@.got.plt)
   0x000e'3e03, // ld      t3, %pcrel_lo(1b)(t3)
   0x000e'0367, // jalr    t1, t3
   0x0010'0073, // ebreak
 };
 
-constexpr ul32 plt_entry_32[] = {
+static constexpr ul32 plt_entry_32[] = {
   0x0000'0e17, // auipc   t3, %pcrel_hi(function@.got.plt)
   0x000e'2e03, // lw      t3, %pcrel_lo(1b)(t3)
   0x000e'0367, // jalr    t1, t3
@@ -433,7 +433,7 @@ void InputSection<E>::apply_reloc_alloc(Context<E> &ctx, u8 *base) {
       // Rewrite `lw t1, 0(t0)` with `lw t1, 0(x0)` if the address is
       // accessible relative to the zero register because if that's the
       // case, corresponding LUI might have been removed by relaxation.
-      if (int_cast(S + A, 12) == S + A)
+      if (is_int(S + A, 12))
         set_rs1(loc, 0);
       break;
     case R_RISCV_TPREL_HI20:
@@ -457,7 +457,7 @@ void InputSection<E>::apply_reloc_alloc(Context<E> &ctx, u8 *base) {
 
       // Rewrite `lw t1, 0(t0)` with `lw t1, 0(tp)` if the address is
       // directly accessible using tp. tp is x4.
-      if (int_cast(val, 12) == val)
+      if (is_int(val, 12))
         set_rs1(loc, 4);
       break;
     }
@@ -518,7 +518,10 @@ void InputSection<E>::apply_reloc_alloc(Context<E> &ctx, u8 *base) {
 
       switch (rel.r_type) {
       case R_RISCV_TLSDESC_LOAD_LO12:
-        write_itype(loc, sym2.get_tlsdesc_addr(ctx) + A - P);
+        if (sym2.has_tlsdesc(ctx))
+          write_itype(loc, sym2.get_tlsdesc_addr(ctx) + A - P);
+        else
+          *(ul32 *)loc = 0x13; // nop
         break;
       case R_RISCV_TLSDESC_ADD_LO12:
         if (sym2.has_tlsdesc(ctx)) {
@@ -540,7 +543,7 @@ void InputSection<E>::apply_reloc_alloc(Context<E> &ctx, u8 *base) {
           write_itype(loc, sym2.get_gottp_addr(ctx) + A - P);
         } else {
           i64 val = S + A - ctx.tp_addr;
-          if (int_cast(val, 12) == val)
+          if (is_int(val, 12))
             *(ul32 *)loc = 0x513;   // addi a0,zero,<lo12>
           else
             *(ul32 *)loc = 0x50513; // addi a0,a0,<lo12>
@@ -831,6 +834,7 @@ template <>
 void shrink_section(Context<E> &ctx, InputSection<E> &isec) {
   std::span<const ElfRel<E>> rels = isec.get_rels(ctx);
   std::vector<RelocDelta> &deltas = isec.extra.r_deltas;
+  i64 r_delta = 0;
   u8 *buf = (u8 *)isec.contents.data();
 
   // True if we can use 2-byte instructions. This is usually true on
@@ -842,8 +846,8 @@ void shrink_section(Context<E> &ctx, InputSection<E> &isec) {
     Symbol<E> &sym = *isec.file.symbols[r.r_sym];
 
     auto remove = [&](i64 d) {
-      i64 sum = deltas.empty() ? 0 : deltas.back().delta;
-      deltas.push_back(RelocDelta{r.r_offset, sum + d});
+      r_delta += d;
+      deltas.push_back(RelocDelta{r.r_offset, r_delta});
     };
 
     // Handling R_RISCV_ALIGN is mandatory.
@@ -854,7 +858,6 @@ void shrink_section(Context<E> &ctx, InputSection<E> &isec) {
     if (r.r_type == R_RISCV_ALIGN) {
       // The total bytes of NOPs is stored to r_addend, so the next
       // instruction is r_addend away.
-      u64 r_delta = deltas.empty() ? 0 : deltas.back().delta;
       u64 P = isec.get_addr() + r.r_offset - r_delta;
       u64 desired = align_to(P, bit_ceil(r.r_addend));
       u64 actual = P + r.r_addend;
@@ -888,15 +891,15 @@ void shrink_section(Context<E> &ctx, InputSection<E> &isec) {
 
       i64 rd = get_rd(buf + r.r_offset + 4);
 
-      if (use_rvc && rd == 0 && int_cast(dist, 12) == dist) {
+      if (use_rvc && rd == 0 && is_int(dist, 12)) {
         // If rd is x0 and the jump target is within ±2 KiB, we can use
         // C.J, saving 6 bytes.
         remove(6);
-      } else if (use_rvc && !E::is_64 && rd == 1 && int_cast(dist, 12) == dist) {
+      } else if (use_rvc && !E::is_64 && rd == 1 && is_int(dist, 12)) {
         // If rd is x1 and the jump target is within ±2 KiB, we can use
         // C.JAL. This is RV32 only because C.JAL is RV32-only instruction.
         remove(6);
-      } else if (int_cast(dist, 21) == dist) {
+      } else if (is_int(dist, 21)) {
         // If the jump target is within ±1 MiB, we can use JAL.
         remove(4);
       }
@@ -919,10 +922,10 @@ void shrink_section(Context<E> &ctx, InputSection<E> &isec) {
         if (rd == get_rd(buf + r.r_offset + 4)) {
           u64 val = sym.get_addr(ctx) + r.r_addend;
 
-          if (use_rvc && rd != 0 && int_cast(val, 6) == val) {
+          if (use_rvc && rd != 0 && is_int(val, 6)) {
             // Replace AUIPC + LD with C.LI.
             remove(6);
-          } else if (int_cast(val, 12) == val) {
+          } else if (is_int(val, 12)) {
             // Replace AUIPC + LD with ADDI.
             remove(4);
           }
@@ -934,13 +937,12 @@ void shrink_section(Context<E> &ctx, InputSection<E> &isec) {
       u64 val = sym.get_addr(ctx) + r.r_addend;
       i64 rd = get_rd(buf + r.r_offset);
 
-      if (int_cast(val, 12) == val) {
+      if (is_int(val, 12)) {
         // We can replace `lui t0, %hi(foo)` and `add t0, t0, %lo(foo)`
         // instruction pair with `add t0, x0, %lo(foo)` if foo's bits
         // [32:11] are all one or all zero.
         remove(4);
-      } else if (use_rvc && rd != 0 && rd != 2 &&
-                 int_cast(val + 0x800, 18) == val + 0x800) {
+      } else if (use_rvc && rd != 0 && rd != 2 && is_int(val + 0x800, 18)) {
         // If the upper 20 bits can actually be represented in 6 bits,
         // we can use C.LUI instead of LUI.
         remove(2);
@@ -969,7 +971,7 @@ void shrink_section(Context<E> &ctx, InputSection<E> &isec) {
       //
       // Here, we remove `lui` and `add` if the offset is within ±2 KiB.
       if (i64 val = sym.get_addr(ctx) + r.r_addend - ctx.tp_addr;
-          int_cast(val, 12) == val)
+          is_int(val, 12))
         remove(4);
       break;
     case R_RISCV_TLSDESC_HI20:
@@ -988,7 +990,7 @@ void shrink_section(Context<E> &ctx, InputSection<E> &isec) {
         assert(r.r_type == R_RISCV_TLSDESC_ADD_LO12);
         if (!sym2.has_tlsdesc(ctx) && !sym2.has_gottp(ctx))
           if (i64 val = sym2.get_addr(ctx) + rel2.r_addend - ctx.tp_addr;
-              int_cast(val, 12) == val)
+              is_int(val, 12))
             remove(4);
       }
       break;
@@ -996,8 +998,7 @@ void shrink_section(Context<E> &ctx, InputSection<E> &isec) {
     }
   }
 
-  if (!deltas.empty())
-    isec.sh_size -= deltas.back().delta;
+  isec.sh_size -= r_delta;
 }
 
 // ISA name handlers
