@@ -353,6 +353,10 @@ void ObjectFile<E>::initialize_sections(Context<E> &ctx) {
           is_debug_section(shdr, name))
         continue;
 
+      // Ignore section is specified by --discard-section.
+      if (ctx.arg.discard_section.contains(name))
+        continue;
+
       if (name == ".comment" &&
           this->get_string(ctx, shdr).starts_with("rustc "))
         this->is_rust_obj = true;
@@ -444,6 +448,13 @@ void ObjectFile<E>::initialize_sections(Context<E> &ctx) {
       target->relsec_idx = i;
     }
   }
+
+  // Attach .arm.exidx sections to their corresponding sections
+  if constexpr (is_arm32<E>)
+    for (std::unique_ptr<InputSection<E>> &isec : this->sections)
+      if (isec && isec->shdr().sh_type == SHT_ARM_EXIDX)
+        if (InputSection<E> *target = sections[isec->shdr().sh_link].get())
+          target->extra.exidx = isec.get();
 }
 
 // .eh_frame contains data records explaining how to handle exceptions.
