@@ -259,9 +259,14 @@ struct FdeRecord {
 template <typename E>
 struct InputSectionExtras {};
 
+template <typename E> requires is_arm32<E>
+struct InputSectionExtras<E> {
+  InputSection<E> *exidx = nullptr;
+};
+
 struct RelocDelta {
-  u64 offset : 38;
-  i64 delta : 26;
+  u64 offset;
+  i64 delta;
 };
 
 // RISC-V and LoongArch support code-shrinking linker relaxation.
@@ -568,6 +573,9 @@ public:
   std::unique_ptr<RelocSection<E>> reloc_sec;
   std::vector<AbsRel<E>> abs_rels;
   Atomic<u32> sh_flags;
+
+  // Used only by create_output_sections()
+  std::vector<std::vector<InputSection<E> *>> members_vec;
 };
 
 template <typename E>
@@ -1290,12 +1298,12 @@ public:
 
   MergedSection<E> &parent;
   std::vector<SectionFragment<E> *> fragments;
+  u8 p2align = 0;
 
 private:
   std::unique_ptr<InputSection<E>> section;
   std::vector<u32> frag_offsets;
   std::vector<u32> hashes;
-  u8 p2align = 0;
 };
 
 // InputFile is the base class of ObjectFile and SharedFile.
@@ -2041,6 +2049,7 @@ struct Context {
     std::optional<std::vector<Symbol<E> *>> retain_symbols_file;
     std::unordered_map<std::string_view, u64> section_align;
     std::unordered_map<std::string_view, u64> section_start;
+    std::unordered_set<std::string_view> discard_section;
     std::unordered_set<std::string_view> ignore_ir_file;
     std::unordered_set<std::string_view> wrap;
     std::vector<SectionOrder> section_order;
@@ -2230,7 +2239,8 @@ enum {
   NO_OPD = 1 << 1, // Request an address other than .opd (PPC64V1 only)
 };
 
-// Symbol class represents a defined symbol.
+// Symbol class represents a symbol. For each unique symbol name, we
+// create one instance of Symbol.
 //
 // A symbol has not only one but several different addresses if it
 // has PLT or GOT entries. This class provides various functions to
