@@ -9,10 +9,10 @@ int main() {
 EOF
 
 $CC -B. -o $t/exe1 $t/a.o -Wl,-package-metadata='{"foo":"bar"}'
-readelf -x .note.package $t/exe1 | grep -Fq '{"foo":"bar"}'
+readelf -x .note.package $t/exe1 | grep -F '{"foo":"bar"}'
 
 $CC -B. -o $t/exe2 $t/a.o -Wl,--package-metadata='%7B%22foo%22%3A%22bar%22%7D'
-readelf -x .note.package $t/exe2 | grep -Fq '{"foo":"bar"}'
+readelf -x .note.package $t/exe2 | grep -F '{"foo":"bar"}'
 
-! $CC -B. -o $t/exe3 $t/a.o -Wl,--package-metadata='foo%x' >& $t/log
-grep -q 'invalid string: foo%x' $t/log
+not $CC -B. -o $t/exe3 $t/a.o -Wl,--package-metadata='foo%x' |&
+  grep 'invalid string: foo%x'
